Set Implicit Arguments.
Require Import LibCore Shared CFHeaps.


(** [hextract] *)

Lemma hextract_demo_1 : forall n J H2 H3 H4 H',
  H4 \* (Hexists y, \[y = 2]) \* 
   (Hexists x, \[n = x + x] \* (J x \* x ~> Id 3) \* H2) \* H3 ==> H'.
Proof using.
  intros. dup 3.
  (* details *)
  hextract_setup tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  try hextract_step tt.
  hextract_cleanup tt.
  intros.
  skip. (* demo *)
  (* short *)
  hextract. skip. (* demo *)
  (* if needed *)
  hextract_if_needed tt. intros.
  try hextract_if_needed tt. skip. (* demo *)
Admitted. (* demo *)

Lemma hextract_demo_2 : forall H1 H' (G:Prop),
  (forall H2, H1 \* (Hexists y, \[y = 2]) \* H2 ==> H' -> G) ->
  G.
Proof using.
  introv W. eapply W. dup.
  (* details *)
  hextract_setup tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  hextract_step tt.
  try hextract_step tt.
  hextract_cleanup tt.
  intros y Hy.
  skip. (* demo *)
  (* short *)
  hextract. skip. (* demo *)
Admitted. (* demo *)

(** [hsimpl] *)

Lemma hsimpl_demo_1 : forall H1 H2 H3 H4 H5,
  H1 \* H2 \* H3 \* H4 ==> H4 \* H3 \* H5 \* H2.
Proof using.
  intros. dup. 
  (* details *)
  hsimpl_setup tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  try hsimpl_step tt.
  hsimpl_cleanup tt.
  skip. (* demo *)
  (* short *)
  hsimpl. skip. (* demo *)
Admitted. (* demo *)

Lemma hsimpl_demo_2 : forall (l1 l2:loc) S1 S2 H1 H2 H3 H',
  (forall X S1' HG, X ==> H1 \* l1 ~> S1' \* H2 \* HG -> X ==> H') ->
  H1 \* l1 ~> S1 \* l2 ~> S2 \* H3 ==> H'.
Proof using.
  intros. dup.
  (* details *) 
  eapply H.
  hsimpl_setup tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  try hsimpl_step tt.
  hsimpl_cleanup tt.
  skip. (* demo *)
  (* short *)
  eapply H.
  hsimpl. skip. (* demo *)
Admitted. (* demo *)

Lemma hsimpl_demo_3 : forall (l1 l2:loc) S1 S2 H1 H2 H',
  (forall X S1' HG, X ==> H1 \* l1 ~> S1' \* HG -> HG = HG -> X ==> H') ->
  H1 \* l1 ~> S1 \* l2 ~> S2 \* H2 ==> H'.
Proof using.
  intros. dup.
  (* details *)
  eapply H.  
  hsimpl_setup tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  try hsimpl_step tt.
  hsimpl_cleanup tt.
  skip.
  auto.
  (* short *)
  eapply H.
  hsimpl.
  auto.
Admitted. (* demo *)

Require Import CFPrimSpec.

Lemma hsimpl_demo_4 : forall x y (n m : int) (My:int->int->hprop),
  x ~> Ref Id n \* y ~> Ref Id n ==> x ~> Ref Id m \* y ~> Ref My m.
Proof using.
  intros. hsimpl. skip. (* demo *)
  (* todo: replace equal with ==> *)
Admitted. (* demo *)

Lemma hsimpl_demo_5 : forall n m J H2 H3 H4,
  n = m + m ->
  H2 \* J m \* H3 \* H4 ==> H4 \* (Hexists y, y ~> Id 2) \* (Hexists x, \[n = x + x] \* J x \* H2) \* H3.
Proof using.
  intros. dup.
  (* details *)
  hsimpl_setup tt.
  hsimpl_step tt.
  hsimpl_step tt. 
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  hsimpl_step tt.
  try hsimpl_step tt.
  hsimpl_cleanup tt.
  auto.
  auto.
  auto.
  (* short *)
  (* TEMPORARY this call to [hsimpl] loops,
     and the calls to [auto] above do nothing
  hsimpl.
  auto.
  auto.
  auto.
  *)
Admitted. (* demo *)

(** [hclean] *)

Lemma hclean_demo_1 : forall A (x X:A) H1 H2 H3 B (F:~~B) Q,  
   is_local F ->
   F (H1 \* \[] \* (H2 \* Hexists y:int, \[y = y]) \* x ~> Id X \* H3) Q.
Proof using.
  intros. dup.
  (* details *)
  hclean_setup tt.
  hclean_step tt.
  hclean_step tt.
  hclean_step tt.
  hclean_step tt.
  hclean_step tt.
  hclean_step tt.
  hclean_step tt.
  hclean_step tt.
  hclean_step tt.
  try hclean_step tt.
  hclean_cleanup tt.
  skip. (* demo *)
  (* short *)
  hclean.
  skip. (* demo *)
Admitted. (* demo *)


(** [hsimpl] on nat credits *)

Lemma hsimpl_credits_nat_demo_1 : forall (x y:nat) (H1 H2 H3:hprop) (a b : loc), 
 \[] \* H2 \* \$_nat (x+y)%nat \* a ~> Id b \* H1 ==> H2 \* \$_nat (x) \* H3.
Proof using. intros. hsimpl. omega. demo. Qed.

Lemma hsimpl_credits_nat_demo_2 : forall (x y z:nat) (H1:hprop), 
 \[] \* H1 \* \$_nat (x) ==> H1 \* \$_nat (y) \* \$_nat (z).
Proof using. intros. hsimpl. demo. demo. demo. Qed.

Lemma hsimpl_credits_nat_demo_3 : forall (x y z t:nat) (H1:hprop), 
 (x >= z + t)%nat ->
 \[] \* H1 \* \$_nat x \* \$_nat y ==> H1 \* \$_nat y \* \$_nat z \* \$_nat t.
Proof using. intros. hsimpl. omega. omega. demo. Qed.


(** [hsimpl] on int credits *)

Lemma hsimpl_credits_int_demo_1 : forall (x y:int) (H1 H2 H3:hprop) (a b : loc), 
 x >= 0 -> y >= 0 ->
 \[] \* H2 \* \$ (x+y) \* a ~> Id b \* H1 ==> H2 \* \$ (x) \* H3.
Proof using. intros. hsimpl. math. math. demo. Qed.

Lemma hsimpl_credits_int_demo_2 : forall (x y z:int) (H1:hprop), 
 \[] \* H1 \* \$ (x) ==> H1 \* \$ (y) \* \$ (z).
Proof using. intros. hsimpl. demo. demo. demo. demo. demo. Qed.

Lemma hsimpl_credits_int_demo_3 : forall (x y z t:int) (H1:hprop), 
 (x >= z + t) -> y >= 0 -> z >= 0 -> t >= 0 ->
 \[] \* H1 \* \$ x \* \$ y ==> H1 \* \$ y \* \$ z \* \$ t.
Proof using. intros. hsimpl; try math. demo. Qed.


(** [csimpl] on int credits *)

Require Import CFLibCredits.

Lemma hsimpl_credits_int_demo_list : forall A (L:list A) x (c:nat),
  \$ c \* \$ (c * LibList.length L) \* \$ c ==> 
  \$ c \* \$ (c * LibList.length (x::L)).
Proof using.
  intros. rew_length. 
  hsimpl. (* optional, but having it is more efficient since it cancels 
             immediately the \$ c. *)
  csimpl. (* same as: hsimpl_credits. csimpl. csimpl. *)
  csimpl.
  (* usually garbage collected *)
  rewrite~ credits_int_zero_eq_prove. simpl_credits. math.
Qed.


(* LATER
Lemma hsimpl_credits_nat_demo_list : forall A (L:list A) x (c:nat),
  \$_nat c \* \$_nat (c * LibList.length L) \* \$_nat c ==> 
  \$_nat c \* \$_nat (c * LibList.length (x::L)).
Proof using.
  intros. rew_length.
  dup. credits_join. demo. hsimpl; try mmath.
  admit. (* distrib *)
  rewrite credits_nat_zero_eq_prove. auto. admit. (* distrib *)
Qed.
*)













